from typing import Optional, List, Union

from optionaldict import OptionalDict

from synology_drive_api.utils import form_urlencoded


def color_name_to_id(color_name: str) -> str:
    """
    map color_name to hex color
    :param color_name:
    :return:
    """
    color_id = {
        'gray': '#A0A5AA',
        'red': '#FA8282',
        'orange': '#FA9C3E',
        'yellow': '#F2CA00',
        'green': '#94BF13',
        'blue': '#499DF2',
        'purple': '#A18AE6',
    }
    try:
        return color_id[color_name.lower()]
    except KeyError:
        raise KeyError('Color name error! Use gray/red/orange/yellow/green/blue/purple')


class LabelsMixin:
    """
    Drive labels related function
    """
    # label dict cache
    _label_dict: dict = dict()

    def get_labels(self, name: Optional[str] = None) -> dict:
        """
        Get label name and label id. If name is None, return all labels.
        :param name: label name
        :return: {name: label_id, ...}
        """
        api_name = 'SYNO.SynologyDrive.Labels'
        endpoint = 'entry.cgi'
        params = {'api': api_name, 'version': 1, 'method': 'list'}
        req = self.session.http_get(endpoint, params=params)
        label_items = req['data']['items']

        if not label_items:
            return dict()

        label_dict = {item['name']: item['label_id'] for item in label_items}
        if name is None:
            return label_dict
        else:
            try:
                return {name: label_dict[name]}
            except KeyError:
                raise Exception(f'Label <{name}> does not exist.')

    def create_label(self, name: str, color: str = 'gray', pos: Union[None, str, int] = None) -> dict:
        """
        :param name: label name
        :param color: color name gray/red/orange/yellow/green/blue/purple
        :param pos: label position
        :return:
        """
        color = color_name_to_id(color_name=color)
        if name in self.label_dict:
            raise Exception('Label_name already exists, please use another one!')
        api_name = 'SYNO.SynologyDrive.Labels'
        endpoint = 'entry.cgi'
        params = OptionalDict(api=api_name, version=1, method='create', name=name, color=color, position=pos)
        ret_label = self.session.http_put(endpoint, params=params)
        self.set_label_dict(name, ret_label['data']['label_id'])
        return ret_label

    def delete_label(self, label_name: Optional[str] = None, label_id: Union[None, str, int] = None) -> dict:
        """
        delete label by name or label_id
        :param label_name: label name
        :param label_id: label id
        :return:
        """
        params_count = [label_name, label_id].count(None)
        if params_count != 1:
            raise Exception('Wrong params number. Enter label name or label_id')

        if label_id is None:
            try:
                label_id = self.label_dict[label_name]
            except KeyError:
                raise Exception(f'Label <{label_name}> does not exists!')

        api_name = 'SYNO.SynologyDrive.Labels'
        endpoint = 'entry.cgi'
        params = {'api': api_name, 'version': 1, 'method': 'delete', 'label_id': label_id}
        return self.session.http_delete(endpoint, params=params)

    def manage_path_label(self, action: str, path: Union[str, List[str]],
                          label: Union[str, List[str], List[dict]]) -> dict:
        """
        add label/labels to file/files or folder/folders
        :param action: 'add', 'delete'
        :param path: file/files, folder/folders
        1. '/team-folders/test_drive/SCU285/test.xls', '/mydrive/test_sheet_file.osheet'
        2. '505415003021516807'
        3. ['505415003021516807', '505415003021516817']
        4. ["id:505415003021516807", "id:525657984139799470", "id:525657984810888112"]
        5. ['/team-folders/test_drive/SCU285/test.xls', '/team-folders/test_drive/SCU283/test2.xls']
        :param label: label/lables
        1. label_name
        2. ['label_name_1', 'lable_name_2']
        3. [{"action": "add", "label_id": "15"}, {"action": "add", "label_id": "16"}]
        :return:
        """
        action = action.lower()
        if action not in ('add', 'delete'):
            raise Exception('Wrong action params. Use add or delete.')

        if isinstance(label, list):
            # label = ['label_name_1', 'label_name_2', 'label_name_3', ...]
            if isinstance(label[0], str):
                label = [{'action': action, 'label_id': self.label_dict.get(single_label)} for single_label in label]
            elif isinstance(label[0], dict):
                # [{"action": "add", "label_id": "15"}, {"action": "add", "label_id": "16"}]
                label = label
            else:
                raise Exception('Wrong label params.')
        elif isinstance(label, str):
            label = [{'action': action, 'label_id': self.label_dict.get(label)}]
        else:
            raise Exception('Wrong label params.')

        if isinstance(path, list):
            path = [f"id:{single_path}" if single_path.isdigit() else single_path for single_path in path]
        elif isinstance(path, str):
            if path.isdigit():
                path = [f"id:{path}"]
            else:
                path = [path]
        else:
            raise Exception('Wrong path params')

        api_name = 'SYNO.SynologyDrive.Files'
        endpoint = 'entry.cgi'
        data = {'files': path, 'labels': label, 'api': api_name, 'method': 'label', 'version': '2'}
        urlencoded_data = form_urlencoded(data)
        return self.session.http_post(endpoint, data=urlencoded_data)

    def list_labelled_files(self, label_name=None, label_id=None, limit=1500) -> dict:
        """
        list specific label files
        :param label_name: label name
        :param label_id:
        :param limit: return result count
        :return:
        """
        params_count = [label_name, label_id].count(None)
        if params_count != 1:
            raise Exception('Wrong params number. Enter label name or label_id')

        if label_id is None:
            try:
                label_id = self.label_dict[label_name]
            except KeyError:
                raise Exception(f'Label <{label_name}> does not exists!')

        api_name = 'SYNO.SynologyDrive.Files'
        endpoint = 'entry.cgi'
        data = {'api': api_name, 'version': 2, 'method': 'list_labelled', 'label_id': label_id, 'offset': 0,
                'limit': limit, 'sort_by': 'name', 'sort_direction': 'desc', 'filter': {}}
        urlencoded_data = form_urlencoded(data)
        return self.session.http_post(endpoint, data=urlencoded_data)

    def set_label_dict(self, label_name, label_id):
        self._label_dict[label_name] = label_id

    @property
    def label_dict(self):
        """
        label name id map
        :return:
        """
        # if cache is disabled, get labels from drive server.
        if not self.enable_label_cache:
            self._label_dict = self.get_labels()
            return self._label_dict

        # if cache is enabled and empty, fill up cache.
        if not self._label_dict:
            self._label_dict = self.get_labels()
        return self._label_dict



from synology_drive_api.utils import form_urlencoded


class TasksMixin:
    """
    tasks related function
    """

    def get_task_status(self, task_id: str):
        """
        :param task_id:
        get task status
        :return:
        """
        api_name = 'SYNO.Entry.Request'
        endpoint = 'entry.cgi'
        data = {
            'stop_when_error': False,
            'mode': 'parallel',
            'api': api_name,
            'compound': [{"api": "SYNO.SynologyDrive.Tasks", "method": "get", "version": 1, "task_id": task_id}],
            'method': 'request',
            'version': 1
        }
        urlencoded_data = form_urlencoded(data)
        resp = self.session.http_post(endpoint, data=urlencoded_data)
        return resp



import io
import os
from pathlib import Path
from time import time, sleep
from typing import Optional, Union, BinaryIO

from synology_drive_api.base import SynologyOfficeFileConvertFailed
from synology_drive_api.utils import concat_drive_path
from synology_drive_api.utils import form_urlencoded
from synology_drive_api.utils import deprecate


class FilesMixin:
    """
    file folder related function
    """

    def get_teamfolder_info(self):
        """
        get teamfolder sub_folder info
        :return: {sub_folder_name: folder_id, ...}
        """
        api_name = 'SYNO.SynologyDrive.TeamFolders'
        endpoint = 'entry.cgi'
        params = {'api': api_name, 'version': 1, 'method': 'list', 'filter': {}, 'sort_direction': 'asc',
                  'sort_by': 'owner', 'offset': 0, 'limit': 1000}
        resp = self.session.http_get(endpoint, params=params)

        if not resp['success']:
            raise Exception('Get teamfolder info failed.')

        if resp['data']['total'] == 0:
            return {}

        return {folder_info['name']: folder_info['file_id'] for folder_info in resp['data']['items']}

    def share_file(self, share_path: str):
        """
        TODO share file more detail params
        :param share_path: id:23333333333  or "'team-folders/folder2/'"
        """
        if share_path.isdigit():
            share_path = f"id:{share_path}"
        api_name = 'SYNO.SynologyDrive.AdvanceSharing'
        endpoint = 'entry.cgi'
        params = {'api': api_name, 'version': 1, 'method': 'create', "path": share_path, "role": "editor"}
        re = self.session.http_put(endpoint, params=params)
        params2 = {'api': "SYNO.SynologyDrive.AdvanceSharing", 'sharing_link': re['data']['sharing_link'],
                   "role": "editor",
                   'method': 'update', 'version': 1, 'due_date': 0, 'path': share_path}
        self.session.http_put(endpoint, params=params2)
        return re

    def create_link(self, file_or_folder_path: str):
        """
        create file link
        :param file_or_folder_path: id:23333333333  or "'team-folders/folder2/'"
        :return:
        """
        api_name = 'SYNO.SynologyDrive.Sharing'
        endpoint = 'entry.cgi'
        if file_or_folder_path.isdigit():
            path_params = f"id:{file_or_folder_path}"
        else:
            # add start position /
            path_params = f"/{file_or_folder_path}" if not file_or_folder_path.startswith('/') else file_or_folder_path
        data = {'api': api_name, 'method': 'create_link', 'version': 1, 'path': path_params}
        urlencoded_data = form_urlencoded(data)
        return self.session.http_post(endpoint, data=urlencoded_data)

    def copy(self, source: str, dist: str) -> dict:
        """
        copy file or dir
        :param source : id:23333333333  or "'team-folders/folder2/'"
        :param dist: "'team-folders/folder2/temp.odoc'"
        """
        endpoint = 'entry.cgi'
        distpath, distname = os.path.split(dist)
        api_name = "SYNO.Office.Node"
        params = {'api': api_name, 'version': 2, 'method': 'copy',
                  'to_parent_folder': distpath, 'dry_run': 'true', 'name': distname,
                  'title': distname[:distname.index('.')],
                  'files': f'["{source}"]'}
        return self.session.http_put(endpoint, params=params)

    def list_folder(self, dir_path: str) -> dict:
        """
        :param dir_path: '/team-folders/folder_name/folder_name1' or '430167496067125111'
        :return:
        """
        if dir_path.isdigit():
            dest_path = f"id:{dir_path}"
        else:
            # add start position /
            dest_path = f"/{dir_path}" if not dir_path.startswith('/') else dir_path

        api_name = 'SYNO.SynologyDrive.Files'
        endpoint = 'entry.cgi'
        params = {'api': api_name, 'version': 2, 'method': 'list', 'filter': {}, 'sort_direction': 'asc',
                  'sort_by': 'owner', 'offset': 0, 'limit': 1000, 'path': dest_path}
        return self.session.http_get(endpoint, params=params)

    def create_folder(self, folder_name: str, dest_folder_path: Optional[str] = None) -> dict:
        """
        Create folder in dest_folder, default location is 'mydrive'. If folder in path does not exist
        :param dest_folder_path: 'id:433415919843151874/folder1', '/mydrive/', 'mydrive/', 'team-folder/folder2/'
        :param folder_name: created folder name
        :return:
        """
        display_path = concat_drive_path(dest_folder_path, folder_name)
        api_name = 'SYNO.SynologyDrive.Files'
        endpoint = 'entry.cgi'
        params = {'path': display_path, 'version': 2, 'api': api_name,
                  'type': 'folder', 'conflict_action': 'autorename', 'method': 'create'}
        return self.session.http_put(endpoint, params=params)

    def get_file_or_folder_info(self, file_or_folder_path: str) -> dict:
        """
        :param file_or_folder_path: file_path or file_id "552146100935505098"
        :return:
        """
        if file_or_folder_path.isdigit():
            path_params = f"id:{file_or_folder_path}"
        else:
            # add start position /
            path_params = f"/{file_or_folder_path}" if not file_or_folder_path.startswith('/') else file_or_folder_path

        api_name = 'SYNO.SynologyDrive.Files'
        endpoint = 'entry.cgi'
        data = {'api': api_name, 'method': 'get', 'version': 3, 'path': path_params}
        urlencoded_data = form_urlencoded(data)
        return self.session.http_post(endpoint, data=urlencoded_data)

    def upload_file(self, file: Union[io.BytesIO, BinaryIO], dest_folder_path: Optional[str] = None,
                    conflict_action='version')-> dict:
        """
        upload file to drive
        :param file: binary_file
        :param dest_folder_path: upload folder path
        :param conflict_action: 'autorename' to rename the new, 'version' to rewrite the file.
                                Default is 'version', same as UI default behaviour.
        :return:
        """
        file_name = file.name
        display_path = concat_drive_path(dest_folder_path, file_name)
        api_name = 'SYNO.SynologyDrive.Files'
        endpoint = 'entry.cgi'
        params = {'api': api_name, 'method': 'upload', 'version': 2, 'path': display_path,
                  'type': 'file', 'conflict_action': conflict_action}
        files = {'file': file}
        upload_ret = self.session.http_post(endpoint, params=params, files=files)
        return upload_ret

    def download_file(self, file_path: str) -> io.BytesIO:
        """
        download file from drive
        :param file_path:
        :return:
        """
        ret = self.get_file_or_folder_info(file_path)
        file_name = ret['data']['name']
        if Path(file_name).suffix in ['.osheet', 'odoc']:
            bio_ret_with_name = self.download_synology_office_file(file_path)
        else:
            file_id = ret['data']['file_id']
            api_name = 'SYNO.SynologyDrive.Files'
            endpoint = f'entry.cgi/{file_name}'
            # \42: "
            params = {'api': api_name, 'method': 'download', 'version': 2, 'files': f"[\42id:{file_id}\42]",
                      'force_download': True, 'json_error': True, '_dc': str(time() * 1000)[:13]}
            bio_ret = self.session.http_get(endpoint, params=params, bio=True)
            bio_ret_with_name = io.BytesIO(bio_ret)
            bio_ret_with_name.name = file_name
        return bio_ret_with_name

    def convert_to_online_office(self, file_path: str, delete_original_file=True, conflict_action='autorename'):
        """
        convert file to online synology office file

        :param file_path:
        :param delete_original_file: indicator delete original file or not
        :param conflict_action: 'autorename' to rename the new, 'version' to rewrite the file
                                Default is 'autorename', same as UI default behaviour.
        :return:
        """
        if not file_path.isdigit() and '.' not in file_path:
            raise Exception('file_path should be id or path with file extension, extensions are xlsx, xls, docx etc.')
        ret = self.get_file_or_folder_info(file_path)
        file_name = ret['data']['name']
        file_extension = Path(file_name).suffix
        if file_extension not in ['.xlsx', '.xls', '.docx']:
            raise SynologyOfficeFileConvertFailed('file_path extension error.')
        file_id = ret['data']['file_id']
        api_name = 'SYNO.SynologyDrive.Files'
        endpoint = 'entry.cgi'
        data = {'api': api_name, 'method': 'convert_office', 'version': 2, 'conflict_action': conflict_action,
                'files': f"[\42id:{file_id}\42]"}
        urlencoded_data = form_urlencoded(data)
        ret = self.session.http_post(endpoint, data=urlencoded_data)
        # when finish converting, delete original file
        if delete_original_file:
            # wait for conversion success
            sleep(1)
            task_status = self.get_task_status(ret['data']['async_task_id'])
            if not task_status['data']['result'][0]['success']:
                sleep(1)
                second_counts = 0
                while True:
                    second_counts = second_counts + 1
                    task_status = self.get_task_status(ret['data']['async_task_id'])
                    if task_status['result'][0]['success'] or second_counts >= 15:
                        break
                if second_counts >= 15:
                    raise SynologyOfficeFileConvertFailed('convert progress timeout(>15s)')
            self.delete_path(file_id)
        return ret

    @deprecate(['upload_file', 'convert_to_online_office'])
    def upload_as_synology_office_file(self, file: Union[io.BytesIO, BinaryIO], dest_folder_path: Optional[str] = None,
                                       upload_conflict_action='version', convert_conflict_action='autorename'):
        """
        upload file to drive, and converted to online office file

        :param file: binary_file
        :param dest_folder_path: upload folder path
        :param upload_conflict_action: 'autorename' to rename the new, 'version' to rewrite the file.
                                        Default is 'version', same as UI default behaviour.
        :param convert_conflict_action: 'autorename' to rename the new, 'version' to rewrite the file.
                                        Default is 'autorename', same as UI default behaviour.
        :return:
        """
        upload_ret = self.upload_file(file, dest_folder_path,
                                      conflict_action=upload_conflict_action)
        # if conversion is failed, delete uploaded file.
        try:
            convert_ret = self.convert_to_online_office(upload_ret['data']['file_id'],
                                                        conflict_action=convert_conflict_action)
        except SynologyOfficeFileConvertFailed as e:
            self.delete_path(upload_ret['data']['file_id'])
            raise e
        return convert_ret

    def download_synology_office_file(self, file_path: str) -> io.BytesIO:
        """
        download synology office file as excel or word
        :param file_path: file/folder or file/folder id "552146100935505098"
        :return:
        """
        if not file_path.isdigit() and '.' not in file_path:
            raise Exception('file_path should be id or path with file extension, extensions are osheet or odoc')

        ret = self.get_file_or_folder_info(file_path)
        file_name = ret['data']['name']
        export_end_point = file_name.replace('osheet', 'xlsx').replace('odoc', 'docx')
        file_id = ret['data']['file_id']

        api_name = 'SYNO.Office.Export'
        endpoint = f"entry.cgi/{export_end_point}"
        params = {'api': api_name, 'method': 'download', 'version': 1, 'path': f"id:{file_id}"}
        bio_ret = self.session.http_get(endpoint, params=params, bio=True)
        bio_ret_with_name = io.BytesIO(bio_ret)
        bio_ret_with_name.name = export_end_point
        return bio_ret_with_name

    def rename_path(self, new_name: str, dest_path: str) -> dict:
        """
        rename file or folder
        :param new_name:
        :param dest_path: file/folder or file/folder id "552146100935505098"
        :return:
        """
        if dest_path.isdigit():
            path_params = f"id:{dest_path}"
        else:
            path_params = dest_path

        api_name = 'SYNO.SynologyDrive.Files'
        endpoint = 'entry.cgi'
        data = {'api': api_name, 'method': 'update', 'version': 2, 'path': path_params, 'name': new_name}
        urlencoded_data = form_urlencoded(data)
        return self.session.http_post(endpoint, data=urlencoded_data)

    def move_path(self, ready_for_move_paths: str, dest_folder: str) -> dict:
        """
        # TODO ready_for_move_paths can be list，str.
        move file or folder to another folder
        :param ready_for_move_paths: file/folder name or file/folder id "552146100935505098"
        :param dest_folder: file/folder name or file/folder id "552146100935505098"
        :return:
        """
        if dest_folder.isdigit():
            dest_path = f"id:{dest_folder}"
        else:
            # add start position /
            dest_path = f"/{dest_folder}" if not dest_folder.startswith('/') else dest_folder

        if ready_for_move_paths.isdigit():
            ready_for_move_paths = [f"id:{ready_for_move_paths}"]
        else:
            ready_for_move_paths = f"/{ready_for_move_paths}" if not ready_for_move_paths.startswith(
                '/') else ready_for_move_paths
            ret = self.get_file_or_folder_info(ready_for_move_paths)
            ready_for_move_paths = [f"id:{ret['data']['file_id']}"]

        # ret = self.get_file_or_folder_info(dest_path)
        api_name = 'SYNO.SynologyDrive.Files'
        endpoint = 'entry.cgi'
        data = {'api': api_name,
                'method': 'move',
                'version': 2,
                'files': ready_for_move_paths,
                'to_parent_folder': dest_path,
                'conflict_action': 'autorename'}
        urlencoded_data = form_urlencoded(data)
        return self.session.http_post(endpoint, data=urlencoded_data)

    def delete_path(self, dest_path: str) -> dict:
        """
        delete file or folder, it's a async task.
        :param dest_path: file/folder or file/folder id "552146100935505098"
        :return:
        """
        ret = self.get_file_or_folder_info(dest_path)
        api_name = 'SYNO.SynologyDrive.Files'
        endpoint = 'entry.cgi'
        data = {'api': api_name, 'method': 'delete', 'version': 2, 'files': [f"id:{ret['data']['file_id']}"],
                'permanent': 'false', 'revisions': ret['data']['revisions']}
        urlencoded_data = form_urlencoded(data)
        return self.session.http_post(endpoint, data=urlencoded_data)



import urllib.parse
import warnings
import functools

import simplejson as json
from selenium import webdriver


def get_local_ip_by_quick_connect_id(q_id):
    """
    # if nas local ip changed, get ip according to quick_connect
    # see https://quickconnect.to/
    :param q_id: QuickConnect ID
    :return:
    """
    url = f"https://{q_id}.quickconnect.to/"
    driver = webdriver.Chrome()
    driver.implicitly_wait(20)
    driver.get(url)
    driver.implicitly_wait(20)
    cookie = driver.get_cookies()
    dict1 = cookie[2]
    value1 = dict1['value']
    return value1.split('ipv4.')[1].split('.wan')[0]


def form_urlencoded(data: dict) -> str:
    """
    Generate urlencoded data for body data
    :param data: ready for post data
    :return: return form data
    """
    data_list = []
    for key, value in data.items():
        value_encode = urllib.parse.quote(json.dumps(value), safe='') if not isinstance(value, str) else value
        # [key, '=', value_encode]
        data_list.append(f"{key}={value_encode}")
    urlencoded_data = '&'.join(data_list)
    return urlencoded_data


def concat_drive_path(dest_path: str, end_point: str, default_folder: str = 'mydrive') -> str:
    """
    Generate file path
    :param dest_path: parent path
    :param end_point: file_name or folder_name
    :param default_folder: if dest_path is None, use your drive home folder instead
    :return:
    """
    if dest_path is None:
        display_path = f"/{default_folder}/{end_point}"
    elif dest_path.startswith('id'):
        display_path = f"{dest_path}/folder_name"
    else:
        # add start position /
        dest_path = f"/{dest_path}" if not dest_path.startswith('/') else dest_path
        # add end position /
        dest_path = f"{dest_path}/" if not dest_path.endswith('/') else dest_path
        display_path = f"{dest_path}{end_point}"
    return display_path


def deprecate(alt_func_names):
    def outer_wrapper(f):
        @functools.wraps(f)
        def inner(*args, **kwargs):
            hint = ', '.join(alt_func_names)
            warnings.warn(
                f'[Deprecation] {f.__name__} will be deprecated in the future.'
                f'Substitutions are: {hint}',
                FutureWarning
            )
            return f(*args, **kwargs)
        return inner
    return outer_wrapper



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['synology_drive_api']

package_data = \
{'': ['*']}

install_requires = \
['optionaldict>=0.1.1,<0.2.0',
 'requests>=2.22.0,<3.0.0',
 'selenium>=3.141.0,<4.0.0',
 'simplejson>=3.17.0,<4.0.0']

setup_kwargs = {
    'name': 'synology-drive-api',
    'version': '1.0.13',
    'description': 'synology drive api python wrapper',
    'long_description': '# Synology Drive API\n[![Downloads](https://static.pepy.tech/personalized-badge/synology-drive-api?period=total&units=international_system&left_color=grey&right_color=red&left_text=Downloads)](https://pepy.tech/project/synology-drive-api)\n![Python](https://img.shields.io/badge/python-v3.7+-blue.svg)\n[![License](https://img.shields.io/badge/license-MIT-gree.svg)](https://opensource.org/licenses/MIT)\n![Contributions welcome](https://img.shields.io/badge/contributions-welcome-orange.svg)\n\n`synology-drive-api` is inspired by  [synology-api](https://github.com/N4S4/synology-api). \nThis repo is aimed at providing **Synology drive** api wrapper and related helper functions. It helps you manage your files/folders/labels in synology drive.  By means of Synology Office, you can edit spreadsheet on Drive and use this api wrapper read spreadsheet. It supports Python 3.7+.\n\n## Installation\n```bash\npip install synology-drive-api\n```\n\n## Get login session\n\nYou can access drive by IP, drive domain or nas domain + drive path. \n\n> Synology Drive allows same user with multiple login session. if you need multiple login session and label functions, disable label cache.\n\n```python\nfrom synology_drive_api.drive import SynologyDrive\n\n# default http port is 5000, https is 5001. \nwith SynologyDrive(NAS_USER, NAS_PASS, NAS_IP) as synd:\n    synd.list_folder(\'/mydrive\')  # write your code here\n# Use specified port\nwith SynologyDrive(NAS_USER, NAS_PASS, NAS_IP, NAS_PORT) as synd:\n    synd.get_file_or_folder_info(path_or_path_id)  # write your code here\n# use http instead of https. https: default is True.\nwith SynologyDrive(NAS_USER, NAS_PASS, NAS_IP, https=False) as synd:\n    synd.create_folder(\'test\', \'team-folder/folder2/\')  # write your code here\n# Enable 2fa.\nwith SynologyDrive(NAS_USER, NAS_PASS, otp_code=\'XXXXXX\') as synd:\n    synd.list_folder(\'/mydrive\')  # write your code here\n# use domain name or name + path access drive\n# Enabled in Application Portal | Application | Drive | General | Enable customized alias\ndrive_path_demo = \'your_nas_domain/drive\'\n# Enabled in Application Portal | Application | Drive | General | Enable customized domain\ndrive_path_demo2 = \'your_drive_domain\'\nwith SynologyDrive(NAS_USER, NAS_PASS, drive_path_demo) as synd:\n    synd.upload_file(file, dest_folder_path=dest_folder_path)  # write your code here\n# disable label cache\nwith SynologyDrive(NAS_USER, NAS_PASS, drive_path_demo, enable_label_cache=False) as synd:\n    synd.list_folder(\'/mydrive\')  # write your code here\n```\nIf you use dsm 7, default dsm_version is \'6\'.  \n```python\nfrom synology_drive_api.drive import SynologyDrive\n\n# default http port is 5000, https is 5001. \nwith SynologyDrive(NAS_USER, NAS_PASS, NAS_IP, dsm_version=\'7\') as synd:\n   synd.download_file(\'/mydrive/test.osheet\')  # write your code here\n```\n## Manage labels\n\nSynology drive thinks labels need to belong to single user. **If you want share labels between users, you should have access to these user accounts.** Another solution is creating a *tool user*.\n\nSynology drive search function provide label union search rather than intersection search. **If you need label intersection search, combine them into one label.**\n\n### Get label info\n\n```python\n# get single label info\nsynd.get_labels(\'your_label_name\')\n# get all labels info\nsynd.get_labels()\n```\n\n### Create/delete label\n\nLabel name is unique in drive.\n\n```python\n# create a label, color name: gray/red/orange/yellow/green/blue/purple.\n# default color is gray, default position is end of labels. 0 is first position.\nret = synd.create_label(\'your_label_name\', color=\'orange\', pos=0)\n# delete label by name/id.\nret = synd.delete_label(\'your_label_name\')\nret = synd.delete_label(label_id=419)\n```\n\n### Add/delete path label\n\n```python\n# acition：add, delete\nsynd.manage_path_label(action, path, label)\n```\n\npath examples:\n\n```python\n1. \'/team-folders/test_drive/SCU285/test.xls\', \'/mydrive/test_sheet_file.osheet\'\n2. \'505415003021516807\'\n3. [\'505415003021516807\', \'505415003021516817\']\n4. ["id:505415003021516807", "id:525657984139799470", "id:525657984810888112"]\n5. [\'/team-folders/test_drive/SCU285/test.xls\', \'/team-folders/test_drive/SCU283/test2.xls\']\n```\n\nlabel examples:\n\n```python\n1. \'label_name\'\n2. [\'label_name_1\', \'lable_name_2\']\n3. [{"action": "add", "label_id": "15"}, {"action": "add", "label_id": "16"}]\n```\n\n### List labelled files\n\nFilter files or folders by single label. If you want to use label union search, use search functions (todo).\n\n```python\nsynd.list_labelled_file(label_name=\'your_label_name\')\n```\n\n## Manage File/Folder\n\n>Team folder start with `/team-folders/`, Private folder start with `/mydrive/`\n\n### List TeamFolder\n\nTeamfolder  is virtual parent folder of shared folders in Synology drive. When you login in Drive, you can see your authorized shared folder.\n\n```python\nsynd.get_teamfolder_info()\n# {sub_folder_name: folder_id, ...}\n```\n\n### List Folder\n\nList Folder or files info of a folder\n\n```python\nsynd.list_folder(\'/mydrive\')\n```\n\n### Get specific folder or file info\n\nGet folder or file info such as created time.\n\n```python\n# file_path or file_id "552146100935505098"\nsynd.get_file_or_folder_info(path_or_path_id)\n```\n\n### Create Folder\n\n```python\n# create folder in your private folder\nsynd.create_folder(folder_name)\n# create folder in dest folder\nsynd.create_folder(\'test\', \'team-folder/folder2/\')\n```\n\n### Upload file\n\nYou don\'t need create folder subfolder before uploading your file.\n\n```python\n# prepare your file\nfile = io.BytesIO(mail_attachment[\'file\'])\n# add a file name to file\nfile.name = strip_file_name(mail_attachment[\'name\'])\nret_upload = synd.upload_file(file, dest_folder_path=dest_folder_path)\n# upload to your private folder\nret_upload = synd.upload_file(file)\n# custom conflict_action: \'version\' to rewrite, \'autorename\' to rename. Default: \'version\'\nret_upload = synd.upload_file(file, dest_folder_path=dest_folder_path, conflict_action=\'version\')\n```\n\nYou can upload xlsx or docx as synology office file.\n\n**[\\*\\*Deprecation hint\\*\\*]** This API will be deprecated in the future. It\'s recommended to call `upload_file` and `convert_to_online_office` by yourself.\n\n``` python\n# custom upload_conflict_action for upload: \'version\' to rewrite, \'autorename\' to rename. Default: \'version\'\n# custom convert_conflict_action for convert: \'version\' to rewrite, \'autorename\' to rename. Default: \'autorename\'\nwith open(\'test.xlsx\', \'rb\') as file:\n    nas_client.upload_as_synology_office_file(file, \'/mydrive/\')\n```\n\n### Convert to online office\n\nTransform docx/xlsx/pptx to Synology online office file.\n\n```python\n# If delete_original_file is True, origin docx/xlsx/pptx will be deleted after transformed. Default: True\n# custom conflict_action: \'version\' to rewrite, \'autorename\' to rename. Default: \'autorename\'\nret_convert = synd.convert_to_online_office(dest_file_path,\n                                            delete_original_file=True,\n                                            conflict_action=\'autorename\')\n```\n\n\n\n### Download file\n\nNew: Support osheet and odoc extensions.\n```python\nfile_name = \'test.osheet\'\nbio = synd.download_file(f\'/mydrive/{file_name}\')\nwith open(file_name, \'wb\') as f:\n    f.write(bio)\n```\n\n### Download Synology office file\n\n```python\nimport pandas as pd\n\n# download osheet as xlsx and read into pandas dataframe.\nbio = synd.download_synology_office_file(\'/mydrive/test.osheet\')  # or\nbio = synd.download_file(\'/mydrive/test.osheet\')\npd.read_excel(bio, sheet_name=None)\n\n# dowloand odoc as docx\nbio = synd.download_synology_office_file(\'/mydrive/test.odoc\')\nwith open(\'test.docx\', \'wb\') as f:\n    f.write(bio)\n```\n\n### Delete file or folder\n\nDelete file or folder is  an async task.\n\n```python\nsynd.delete_path(\'/mydrive/abc_folder\')\nsynd.delete_path(\'598184594644187768\')\n```\n\n### Rename file or folder\n\n```python\n# Rename file \'/mydrive/H3_AP201812091265503218_1.pdf\' to \'/mydrive/new.pdf\'\nsynd.rename_path(\'new.pdf\', \'/mydrive/H3_AP201812091265503218_1.pdf\')\n# Rename folder \'/mydrive/test_folder\' to \'/mydrive/abc_folder\'\nsynd.rename_path(\'abc_folder\', \'/mydrive/test_folder\')\n```\n\n### Share file or folder\n\nGet unique file url.\n\n```python\nsynd.create_link(\'team-folders/operation/H3_AP201812091265503218_1.pdf\')\n```\n',
    'author': 'zbjdonald',
    'author_email': 'zbjdonald@qq.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/zbjdonald/synology-drive-api',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.7,<4.0',
}


setup(**setup_kwargs)



from typing import Optional, Union

from synology_drive_api.base import SynologySession
from synology_drive_api.files import FilesMixin
from synology_drive_api.labels import LabelsMixin
from synology_drive_api.tasks import TasksMixin


class SynologyDrive(LabelsMixin, FilesMixin, TasksMixin):
    # synology login session
    session: SynologySession
    # if you need multiple login session and label functions, disable label cache. Default behavior is enabling cache.
    enable_label_cache: bool

    def __init__(self,
                 username: str,
                 password: str,
                 ip_address: Optional[str] = None,
                 port: Union[None, str, int] = None,
                 nas_domain: Optional[str] = None,
                 https: bool = True,
                 enable_label_cache: bool = True,
                 dsm_version: str = '6',
                 max_retry: int = 2,
                 otp_code: Optional[str] = None) -> None:
        self.session = SynologySession(username, password, ip_address, port, nas_domain, https, dsm_version, max_retry,
                                       otp_code)
        self.enable_label_cache = enable_label_cache

    def __enter__(self):
        self.login()
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.logout()

    def login(self):
        return self.session.login('SynologyDrive')

    def logout(self):
        return self.session.logout('SynologyDrive')

    def get_info(self):
        api_name = 'SYNO.SynologyDrive.Info'
        info = self.session.get_api_list(api_name)
        endpoint = info['path']
        params = {'api': api_name, 'version': info['maxVersion'], 'method': 'get', '_sid': self.session.sid}
        return self.session.http_get(endpoint, params=params)



import functools
from http import cookiejar
from time import sleep
from urllib.parse import urlparse

import requests
import simplejson as json
import urllib3
from typing import Optional, Union

# Used for verify=False in requests
urllib3.disable_warnings(urllib3.exceptions.InsecureRequestWarning)


class BlockAll(cookiejar.CookiePolicy):
    """
    drop cookies from session object
    """
    return_ok = set_ok = domain_return_ok = path_return_ok = lambda self, *args, **kwargs: False
    netscape = True
    rfc2965 = hide_cookie2 = False


class SynologyException(requests.RequestException):

    def __init__(self, code=-1, message=None, *args, **kwargs):
        super(SynologyException, self).__init__(*args, **kwargs)
        self.code = code
        self.message = message

    def __str__(self):
        return self.response.text

    __repr__ = __str__


class SynologyOfficeFileConvertFailed(Exception):
    pass


def concat_nas_address(ip_address: Optional[str] = None,
                       port: Union[None, str, int] = None,
                       drive_prefix: Optional[str] = None,
                       https: bool = True) -> str:
    """
    :param ip_address: such as 192.168.1.51
    :param port: port number
    :param drive_prefix: such as drive.xxxx.com, hehe.com/drive
    :param https:
    :return:
    """
    scheme = 'https://' if https else 'http://'
    if port is None:
        # https default port 5001, http default port 5000
        port = '5001' if https else '5000'
    else:
        # int => str
        port = str(port)

    if ip_address is None:
        assert isinstance(drive_prefix, str)
        nas_address = ':'.join(filter(None, [drive_prefix]))
    else:
        assert drive_prefix is None
        nas_address = ':'.join(filter(None, [ip_address, port]))

    return f"{scheme}{nas_address}"


def add_sid_token(reqs_data: dict, sid: str) -> dict:
    """
    add sid token in requests params
    :param reqs_data: requests kwargs
    :param sid: sid token
    :return:
    """
    if 'params' not in reqs_data:
        reqs_data['params'] = {}

    # api params may in forms sometimes
    if 'api' not in reqs_data['params'] and 'api=' not in reqs_data.get('data', ''):
        return reqs_data

    # login logout don't need sid
    if reqs_data['params'].get('api') != 'SYNO.API.Auth':
        reqs_data['params']['_sid'] = sid
    return reqs_data


def raise_synology_exception(resp, bio_exist: bool = False) -> None:
    """
    :param resp:
    :param bio_exist: indicate response contains binary object
    :return:
    """
    # handle 404, 400, ..., and synology server error code
    try:
        resp.raise_for_status()
    except requests.RequestException as reqe:
        code = -1
        message = None
        if reqe.response.text:
            try:
                err_info = json.loads(reqe.response.text)
                code = err_info['code']
                message = err_info['message']
            except (ValueError, KeyError):
                pass
        raise SynologyException(
            code=code,
            message=message,
            request=reqe.request,
            response=reqe.response
        )

    if resp.status_code == 200 and not bio_exist:
        result = resp.json() if resp.text else {}
        if not result['success']:
            raise SynologyException(
                code=result['error']['code'],
                # sometimes there is no 'errors' key
                message=result['error'].get('errors') if result['error'].get('errors') else result['error'],
                request=resp.request,
                response=resp
            )


class SynologySession:
    """
    Synology App base class
    """
    _username: str
    _password: str
    _otp_code: str
    # api base url
    _base_url: str
    # sid token
    _sid: Optional[str] = None
    req_session: requests.Session = requests.Session()
    _session_expire: bool = True
    # dsm version, used for login api version
    dsm_version: str = '6'
    max_retry: int = 2

    def __init__(self,
                 username: str,
                 password: str,
                 ip_address: Optional[str] = None,
                 port: Optional[int] = None,
                 nas_domain: Optional[str] = None,
                 https: Optional[bool] = True,
                 dsm_version: str = '6',
                 max_retry: int = 2,
                 otp_code: Optional[str] = None) -> None:
        assert dsm_version in ('6', '7'), "dsm_version should be either '6' or '7'."

        nas_address = concat_nas_address(ip_address, port, nas_domain, https)
        self._username = username
        self._password = password
        self._otp_code = otp_code
        self.dsm_version = dsm_version
        self._base_url = f"{nas_address}/webapi/"
        self.req_session.cookies.set_policy(BlockAll())
        self.max_retries = max_retry

    def _request(self, method: str, endpoint: str, **kwargs):
        """
        :param method: get, post, put, delete
        :param endpoint: api endpoint
        :param kwargs:
        :return:
        """
        if not endpoint.startswith(('http://', 'https://')):
            api_base_url = kwargs.pop('api_base_url', self._base_url)
            url = f"{api_base_url}{endpoint}"
        else:
            url = endpoint

        kwargs = add_sid_token(kwargs, self._sid)

        if isinstance(kwargs.get('data', ''), dict):
            body = json.dumps(
                kwargs['data'],
                ensure_ascii=False,
            )
            body = body.encode('utf-8')
            kwargs['data'] = body
        elif isinstance(kwargs.get('data'), str):
            kwargs['data'] = kwargs['data'].encode('utf-8')

        parsed_url = urlparse(url)
        # Allow url pattern: https://192.168.1.58:133/webapi
        if parsed_url.scheme == 'https' and parsed_url.netloc.count('.') >= 3:
            # if scheme is https:// and url contains ip,
            # return true indicate adding verify=False to requests.
            kwargs['verify'] = False
        bio_flag = kwargs.pop('bio') if 'bio' in kwargs else None
        if not self.max_retries:
            res = self.req_session.request(
                method=method,
                url=url,
                **kwargs
            )
            raise_synology_exception(res, bio_exist=bio_flag)
        else:
            for retry in range(self.max_retries):
                try:
                    res = self.req_session.request(
                        method=method,
                        url=url,
                        **kwargs
                    )
                    raise_synology_exception(res, bio_exist=bio_flag)
                    break
                except SynologyException as e:
                    # retry
                    # 105: permission denied by anonymous
                    # 1003 1002: get file information failed
                    if e.code in (105, 1003, 1002):
                        sleep(0.5 * (retry + 1))
                        if retry == self.max_retries - 1:
                            raise e
                        else:
                            continue
                    else:
                        raise e
        if bio_flag:
            return res.content
        result = res.json() if res.text else {}
        return result

    def http_get(self, endpoint: str, **kwargs):
        return self._request('get', endpoint, **kwargs)

    def http_post(self, endpoint: str, **kwargs):
        return self._request('post', endpoint, **kwargs)

    def http_put(self, endpoint: str, **kwargs):
        return self._request('put', endpoint, **kwargs)

    def http_delete(self, endpoint: str, **kwargs):
        return self._request('delete', endpoint, **kwargs)

    def login(self, application: str):
        endpoint = 'auth.cgi'
        login_api_version = '2' if self.dsm_version == '6' else '3'
        params = {'api': 'SYNO.API.Auth', 'version': login_api_version, 'method': 'login', 'account': self._username,
                  'passwd': self._password, 'session': application, 'format': 'cookie'}
        if self._otp_code is not None:
            params['otp_code'] = self._otp_code

        if not self._session_expire:
            if self._sid is not None:
                self._session_expire = False
                return 'User already logged'
        else:
            resp = self.http_get(
                endpoint,
                params=params
            )
            self._sid = resp['data']['sid']
            self._session_expire = False
            return 'User logging... New session started!'

    def logout(self, application: str):
        endpoint = 'auth.cgi'
        logout_api_version = '2' if self.dsm_version == '6' else '3'
        params = {'api': 'SYNO.API.Auth', 'version': logout_api_version, 'method': 'logout', 'session': application}
        resp = self.http_get(
            endpoint,
            params=params
        )
        if resp['success'] is True:
            self._session_expire = True
            self._sid = None
            return 'Logged out'
        else:
            self._session_expire = True
            self._sid = None
            return 'No valid session is open'

    @functools.lru_cache()
    def get_api_list(self, app=None):
        endpoint = 'query.cgi'
        params = {'api': 'SYNO.API.Info', 'version': '1', 'method': 'query', 'query': 'all'}
        resp = self.http_get(
            endpoint,
            params=params
        )
        if app is not None:
            for key in resp['data']:
                if app.lower() in key.lower():
                    return resp['data'][key]
        else:
            return resp['data']

    @property
    def sid(self):
        return self._sid
